# 函数编写指南

`@cloudbase/functions-framework` 是一个函数式代码运行框架，本文档主要说明如何编写基于 `@cloudbase/functions-framework` 框架的函数代码。
例如，如何进行错误处理，以及函数结构、出入参等基本信息，并提供一系列的示例以供参考。

## 快速开始

### 第一步：编写函数

1. 在本地创建一个空的文件夹，作为项目的根目录。
2. 新建 `index.js` 文件，作为函数的入口文件。
3. 在`index.js` 文件中，填写以下内容：

```js
exports.main = function (event, context) {
  const { httpContext } = context
  const { url, httpMethod } = httpContext
  return `[${httpMethod}][${url}] Hello world!`
}
```

### 第二步：部署函数

可以部署到任意支持运行 `Node.js` 的环境中。

### 第三步：调用函数

当部署函数之后，即可通过 HTTP 请求来调用函数。

## 了解函数

### 函数的输入参数 `event` 和 `context`

函数的基本结构如下：

```js
exports.main = function (event, context) {
  // 函数逻辑
}
```

或

```js
exports.main = async function (event, context) {
  // 函数逻辑
}
```

该文件导出一个 `main` 函数，该函数即为函数的入口方法，当请求到达时，该方法会作为代码执行的起点。该方法接受两个固定的参数 `event` 和 `context`，分别代表函数的输入和上下文。

其中，`event` 为函数的触发事件，在当前的 HTTP 访问场景下，可以认为是 HTTP 请求体，例如 `POST` 请求的 `body`，`multipart/form-data` 请求的 `form-data`，`PUT` 请求所传输的二进制文件等。如果 HTTP 请求没有传递请求体，`event` 将会是一个空对象。

`context` 为函数执行的上下文信息，包含以下属性和方法：

| 名称          | 类型                      | 含义                                                     |
| ------------- | ------------------------- | -------------------------------------------------------- |
| eventID       | `string`                  | 事件唯一标识，用于关联上下文                                 |
| eventType     | `string`                  | 事件类型，当前固定为 `http`                                |
| timestamp     | `number`                  | 请求时间戳                                               |
| httpContext   | `httpBasis`               | HTTP 请求的相关信息                                      |
| tcbContext    | `Record<string, unknown>` | 云开发相关上下文，调用时注入（仅在云开发平台运行时有值）        |
| sse()         | `ISeverSentEvent`         | 返回用于发送 `Server-Sent Event` 格式响应的对象            |

`httpBasis` 定义如下：

| 名称       | 类型                  | 含义                           |
| ---------- | --------------------- | ------------------------------ |
| url        | `string`              | 本次请求的完整 URL             |
| httpMethod | `string`              | 本次请求的 HTTP 方法，如 `GET` |
| headers    | `IncomingHttpHeaders` | 本次请求的 HTTP 头部           |

云函数支持 `Server-Sent Event` 格式的响应，通过调用 `sse()` 即可以切换到 `SSE` 模式并得到 `ISeverSentEvent` 实例。

```ts
interface ISeverSentEvent {
  setEncoder(encoder: TextEncoder): void;
  send(event: SseEvent): boolean;
  end(msg: SseEvent): void;
}

interface SseEvent<T = string | Record<string, unknown>> {
  data: T;
  comment?: string;
  event?: string;
  id?: string;
  retry?: number;
}
```

可以通过 `send()` 方法来发送 SSE 响应，例如：

```js
const { sse } = context
sse.send({
  data: "Hello world!"
})
```

当函数执行完毕（即从入口函数返回）时，返回值将会被直接返回给客户端，不会进行额外的序列化处理。

### 函数的输出（即返回值）

数执行最后的返回值将作为最终返回值返回给客户端，此时，HTTP 将使用默认的状态码（如正常返回时`200`、未带有响应体时`204`、出错时`500`）和默认的响应头返回给客户端。

函数支持 同步、异步 两种形式的写法，函数支持 `普通类型` | `Promise 类型` 的返回值：

```ts
type ResponseBodyT = string | number | boolean | object | null | undefined | Buffer | Stream | IntegrationResponse
FunctionResponse | Promise<FunctionResponse>
```

通过 `Buffer | Stream` 类型的返回值，可以直接返回二进制数据，实现下载文件的能力。

#### 集成响应

通过集成响应，可以自定义 HTTP 响应的结构，包括响应状态码、头部字段等，可以采用集成响应的返回结构。集成响应的定义如下：

```ts
interface IntegrationResponse {
  statusCode: number
  headers: { [key: string]: string | string[] }
  body: unknown
}
```

`statusCode` 为 HTTP 响应状态码，`headers` 为自定义的 HTTP 响应头部，`body` 为 HTTP 响应体。例如返回下面的集成响应：

```js
exports.main = function(event, context) {
  return {
    statusCode: 200,
    headers: {
      "Content-Type": "text/html"
    },
    body: {
      message: "<h1>Hello world!</h1>"
    }
  }
}
```

将可以在浏览器中渲染出 HTML。

#### 错误处理

函数运行过程中可能或抛出异常，异常整体可以分为以下几类：

1. 不可恢复的错误：如启动时未加载成功代码、Out of Memory(OOM)、堆栈溢出，这类错误会导致函数无法启动或从运行中退出，此类错误是严重错误，会导致函数实例重启，应当尽量规避；
2. 请求处理过程中的错误，即业务逻辑报错：这类错误通常是可捕获的，函数需要对这些错误进行处理，记录相关日志并返回相关信息给客户端。如果用户代码未捕获相关异常，函数框架将会捕获该错误，记录相关日志并返回框架定义的错误信息给客户端；
3. 未被捕获的异常 `UncaughtException` 和 `UnhandledRejection`：通常发生在异步代码逻辑中，因为是未捕获异常，业务上是感知不到的，这类错误会被框架捕获，并打印相关信息到日志中。因为通常发生在异步逻辑中，该错误不会体现在函数返回值中；
4. 函数框架报错：函数框架内部可能发生一些错误，可能是函数框架代码存在 Bug，也可能是函数框架的限制，这类错误会被框架捕获，并打印相关信息到日志中；
5. 网络错误：在客户端往返函数实例的网络链路中也可能出现异常，导致一些错误。该类错误需结合具体部署平台进行排查。

### 函数的模块化和依赖安装

如通常的 `Node.js` 代码一样，函数也可以通过 `Javascript` 模块化方式来组织代码。对于遵循 `CommonJS` 规范的模块，可以通过 `require` 来引入依赖。例如：

```js
// index.js
const { add } = require('./sum')

exports.main = function(event, context) {
  return add(1, 2)
}

// sum.js
exports.add = function(a, b) {
  return a + b
}
```

对使用 `ECMAScript` 模块化（需在 `package.json` 中声明模块化方式）的代码，可以通过 `import` 语句来引入依赖。例如：

```js
// index.mjs
import { add } from './sum.mjs'

export function main(event, context) {
  return add(1, 2)
}

// sum.mjs
export function add(a, b) {
  return a + b
}

// package.json
{
  "name": "example",
  "version": "1.0.0",
  "main": "index.mjs",  // 指定入口文件
  "type": "module",     // 声明模块化方式
  "dependencies": {
  }
}
```

对于外部依赖安装，可在 `package.json` 中声明依赖，函数构建时会根据依赖声明自动安装依赖包。如果目录下存在 `package-lock.json` 等 lock 文件，可以起到锁定依赖版本的作用。

### 函数的触发方式和执行模式

函数本质上是一个运行在 Unix 环境中的 HTTP 服务，函数的执行由 HTTP 请求触发，当长时间没有请求时，如果开启了自动扩缩容，函数服务会自动缩容，避免不必要的资源消耗。

单个函数实例可以并发处理多个请求，多个请求可能被多个实例处理，也可能不是，因此不同的请求之间不应当存在依赖关系，各个请求应保持自身的独立性。

在函数中执行的代码支持 Node.js 运行时支持的能力，包括同步、异步操作、读取环境变量、启用长连接等能力，但函数实例并不是持久化的，其所使用的临时磁盘空间可能会在实例销毁时一同清空，因此不应依赖本地磁盘存储需要持久化的数据。此外，函数所处的操作系统也可能会发生变化，在函数代码中，应当避免对特定操作系统的依赖。

函数是平台分配资源的基本单位，根据函数的 CPU、内存资源使用情况，平台会在满足设定的参数的情况下，动态调度函数实例的数量，以达到性能和资源利用率的平衡。

### 函数日志

在函数代码中打印的日志可以分为两类：

* 一类是 `函数实例级别日志`，即入口函数之外的日志，这部分日志在函数实例启动时、模块被加载时被打印，跟某次具体的请求无关；
* 另一类是 `请求级别日志`，即某次请求触发、代码执行流程进入主入口函数内部之后才会打印的日志，这部分日志可通过某次具体请求 `eventID` 相关联。

`请求级别日志` 从内容上划分，又可以分为以下几类：

1. 每次请求由函数框架自动打印的结构化访问日志（`accesslog`），包括该次请求的基本信息，如处理请求的主机名、访问的 URL、请求方法、响应结果、请求处理耗时等；
2. 函数代码中主动打印的日志，即代码中使用 `console.log` 等方法打印的日志，这部分日志会跟 `eventID` 所关联，有助于排查某次请求的具体执行流程；
3. 函数框架捕获的异常日志，如 `UncaughtException` 和 `UnhandledRejection` ，这部分日志会自动跟 `eventID` 关联，并打印出异常的堆栈信息等便于排查问题。

日志的逻辑级别如下所示：

```text
函数日志
├── 函数实例日志
├── 请求级别日志
│   ├── 访问日志
│   ├── 代码中打印的日志
│   ├── 框架捕获的异常日志
```

例如下面的函数代码：

```js
// 函数实例日志
console.log("func initialization started.")

// 函数实例日志
setTimeout(() => { 
  console.log("timeout log.")
}, 1000)

exports.main = function(event, context) {
  // 请求级别日志
  console.log({ event, context })

  setTimeout(() => { throwReject() }, 1000)
  return "Hello world"
}

async function throwReject() {
  // 这里会触发未捕获的异常，因为没有捕获，会被框架捕获
  // 该异常在 main 中 setTimeout 中抛出，因此不会体现在 main 的返回值中
  // 该异常由 函数框架捕获并打印到日志中
  Promise.reject(new Error('This is an error.'))
}

// 函数实例日志
console.log("func initialization finish.")
```

以上示例函数，在请求到达并处理完毕后，客户端会收到 `Hello world` 响应，1s 后该请求抛出异常被框架捕获，这次请求的访问日志形如：

```json
{
  "@timestamp": "2024-xx-xxTxx:xx:xx.xxxZ", // 日志打印的时间戳
  "startAt": "2024-xx-xxTxx:xx:xx.xxxZ", // 请求开始处理的时间
  "endAt": "2024-xx-xxTxx:xx:xx.xxxZ", // 请求处理结束的时间
  "logType": "accesslog", // 日志类型
  "hostname": "host", // 处理请求的主机名
  "pid": 13506, // 进程 ID
  "version": "x.y.z", // 函数框架版本
  "nodeEnv": "", // Node.js 相关的环境变量
  "eventId": "b0900934-79d7-4441-856f-dd46392a5f91", // 本次请求的 eventID
  "method": "GET", // 请求方法
  "url": "http://127.0.0.1/", // 请求 URL
  "path": "/", // 请求路径
  "status": 200, // 响应状态码
  "httphost": "127.0.0.1", // 请求的主机名
  "ua": "PostmanRuntime/7.39.0", // 请求的 User-Agent
  "connection": "keep-alive",
  "contentType": "application/json", // 请求的 Content-Type
  "clientIp": "::1", // 请求的客户端 IP
  "resSize": 12, // 响应体大小
  "resContentType": "text/plain", // 响应的 Content-Type
  "timeCost": 11, // 本次请求的总耗时
  "userTimeCost": 0 // 从函数主入口开始到从入口返回的耗时
}
```

随后异常会被框架捕获，可通过 `eventID`(`b0900934-79d7-4441-856f-dd46392a5f91`) 关联，并打印出如下堆栈：

```text
Unhandled Rejection for: Error: This is an error.
    at throwReject (/path/to/your/project/index.js:9:18)
    at Timeout._onTimeout (/path/to/your/project/index.js:4:22)
```

如果业务代码按框架定义的错误结构 `{code: string, message?: string}` 返回错误。框架会将该结构中的 `code` 和 `message` 记录到访问日志中，方便排查问题。

```json
{
  "code": "SOME_THING_ERROR",
  "message": "This is a business error.",
}
```

> 注意：日志对于问题排查是非常重要的，建议在函数代码中适当打印日志，以便排查问题。
> 但是打印大量日志将会带来其他问题，如日志消耗磁盘空间可能导致磁盘可用空间不足，日志采集组件采集解析文件消耗较多 CPU，日志上报到存储系统会消耗大量网络带宽，存储大量日志导致成本增加，日志写入和查询性能下降延迟高等诸多问题，因此建议适度打印日志。

## 和云开发·云函数的对比

下表对比了 `云开发·云函数` 和 `基于 functions-framework 函数` 的主要区别：

| 分类              | 云开发·云函数                        | 基于 `functions-framework` 函数                        |
|------------------|------------------------------------|-------------------------------------------------------|
| **部署方式**      | 仅支持部署到云开发云函数平台            | 可以部署到任意支持运行 `Node.js` 的环境中，支持本地运行        |
| **并发模型**      | 单实例单并发                         | 单实例多并发                                             |
| **HPA模式**      | 云函数平台根据并发数量HPA，可预制并发    | 基于部署平台HPA能力，如k8s容器平台可以基于CPU/MEM配置HPA       |
| **内存限制**      | 支持MEM大小限制                      | 通过运行平台控制                                         |
| **CPU限制**      | 不感知CPU核数                        | 感知CPU核数，通过运行平台控制                              |
| **超时控制**      | 支持函数执行超时控制，超时后可中断执行    | 支持函数调用超时控制，超时后函数代码可继续执行                 |
| **调试能力**      | 本地调试较难                         | 框架支持本地运行，支持热重启，调试更容易                      |
| **日志**         | 提供调用记录、高级日志                 | 提供访问日志、高级日志，访问日志面向用户提供更多信息            |
| **依赖安装**      | 支持云端 npm 安装依赖                 | 支持云端 npm、yarn、pnpm 安装依赖                        |
| **运行时版本**    | 平台提供的指定 node.js 运行时版本       | 可使用满足最低要求的任意 node.js 运行时版本                 |
| **触发方式差异**  | 通过云开发SDK、HTTP调用                | 原生HTTP调用，支持获取 HTTP 请求上下文                     |
| **文件上传**     | 不支持文件上传                         | 支持 `form-data` 等 HTTP原生二进制 上传能力               |
| **文件下载**     | 不支持文件下载                         | 支持文件下载                                            |
| **流式响应**     | 不支持                                | 支持 Stream 流式响应                                    |
| **常驻实例**     | "不支持"（不感知函数实例）               | 感知实例概念，实例可以自定义是否常驻                        |
| **预制并发**     | 支持                                 | 无此概念，类似于手动扩容                                  |
| **长连接能力**    | 不支持                               | 支持 SSE、WebSocket 方式的长连接能力                     |
| **连接数据库**    | 基于单实例单并发模型，对数据库长连接不友好  | 跟普通 node.js 程序无区别                               |
| **异步函数**      | 支持，静态配置无法动态切换，限制执行时长   | 支持，函数代码自行控制是否异步执行                          |
| **异步任务**      | 部分版本支持                          | 支持，函数代码自行控制是否异步执行                          |
| **链路差异**      | 链路耗时较长                         | 依赖部署方式，通常链路耗时较短                             |
| **计费方式**      | 计量单位GBs，按量计费                 | 部署平台资源计费方式决定                                   |

## 示例集

### 引入外部模块

在函数中引入外部模块，例如 `lodash`，并使用其提供的方法：

```js
// index.js
const _ = require('lodash')

exports.main = function(event, context) {
  return _.kebabCase('Hello world')
}
```

```json
// package.json
{
  "name": "example",
  "version": "1.0.0",
  "main": "index.js",
  "dependencies": {
    "lodash": "^4.17.21"
  }
}
```

执行该函数后客户端可以收到 `hello-world` 的响应。

### 使用 Server-sent Event 推送消息

```js
exports.main = async function (event, context) {
  // 切换到 SSE 模式
  const sse = context.sse()

  sse.on('close', () => {
    console.log('sse closed')
  })

  // 发送事件到客户端，发送前先检查是否已经关闭，如未关闭可发送
  if (!sse.closed) {
    // 多次发送多个事件
    sse.send({ data: 'No.1 message' })
    sse.send({ data: 'No.2 message with\n\r\r\n\r\r\rtwo lines.' })

    // 单次发送多个事件
    sse.send([
      { data: 'No.1 message' },
      { data: 'No.2 message with\n\r\r\n\r\r\rtwo lines.' }
    ])

    // 以下为发送原始消息的示例
    // 该方式用于扩展 SSE 协议，例如发送其他 Event Field 字段
    // 注意：末尾必须有换行符数据才会立即发送
    sse.send('message: This is a raw message. ')
    sse.send(['message: This is another raw message.\n\n'])

    // 函数执行时间以函数返回时间计算
    // 函数返回后，HTTP 请求处理完成，函数内的异步逻辑继续进行处理，不影响函数返回时间
    // TCP 网络连接依然被 SEE 占用，在 SSE 连接被客户端或服务端关闭之前，可以继续发送消息到客户端
    // SSE 协议已经将 HTTP 转换到长连接模式，需要客户端或服务端在适当的时候主动关闭连接，否则将导致连接一直占用，消耗网络资源
    // 因TCP主动关闭的一方将进入TIME_WAIT 状态，大量 TIME_WAIT 状态的连接将导致网络资源耗尽，无法建立新的连接，所以客户端主动关闭连接更符合最佳实践
    // 因客户端可能并不知晓在什么时间关闭连接，服务端可以发送一个特殊的消息，告诉客户端消息已经结束，可以关闭连接了
    return ''
  }
}
```

### 在函数中路由

可根据 `context` 中获取到的 HTTP 相关路径、query 等信息，实现简单的路由功能：

```js
exports.main = function(event, context) {
  const {  httpContext } = context
  const { url } = httpContext
  const path = new URL(url).pathname

  // 根据访问路径返回不同的内容
  switch (path) {
    case '/':
      return {
        statusCode: 200,
        body: 'Hello world!'
      }
    case '/index.html':
      return {
        statusCode: 200,
        headers: {
        'Content-Type': 'text/html'
        },
        body: '<h1>Hello world!</h1>'
      }
    default:
      return {
        statusCode: 404,
        body: 'Not found'
      }
  }
}
```

### 使用 POST `multipart/form-data` 上传文件

```js
const path = require('path')
const fs = require('fs')

exports.main = async function (event, context) {
  // 从 event.file 属性上获取要保存的文件（与传参对应）
  // event.file 类型为 PersistentFile，见 https://www.npmjs.com/package/formidable#file
  // 如有其他传参，可用 form data 传参中的对应名称 event.[your param name] 获取，如 event.name, event.size 等
  const file = event.file
  // 获取原始文件名
  const fileName = file.originalFilename;
  // 保存文件的目录
  const fileDir = path.join(process.cwd(), 'tmp')
  if (!fs.existsSync(fileDir)) {
    fs.mkdirSync(fileDir)
  }
  const filePath = path.join(fileDir, fileName)
  // 从参数中读取文件流
  const readStream = fs.createReadStream(file.filepath)
  // 尝试保存文件到指定目录
  try {
    await fs.promises.writeFile(filePath, readStream)
    return {
      statusCode: 200,
      body: `File saved to: ${filePath}`
    }
  } catch (error) {
      return {
        statusCode: 500,
        body: `Error saving file: ${error.message}`
      }
  }
}
```

发送上传文件请求：

```sh
curl --location 'url' \
    --form 'file=@file.png'
```

### 使用 PUT 上传文件

```js
const path = require('path')
const fs = require('fs')

exports.main = async function(event, context) {
  const { httpContext } = context
  const { url } = httpContext
  // 从 query 中获取文件名
  const filename = new URL(url).searchParams.get('filename')
  // 保存文件的目录
  const fileDir = path.join(process.cwd(), 'tmp')
  if (!fs.existsSync(fileDir)) {
      fs.mkdirSync(fileDir)
  }
  // 保存文件的路径
  const filePath = path.join(fileDir, filename)

  try {
    // 从 event 中获取文件内容
    const buffer = Buffer.from(event, 'binary')
    await fs.promises.writeFile(filePath, buffer);
    return {
      statusCode: 200,
      body: `File saved to: ${filePath}`,
    };
  } catch (error) {
    return {
      statusCode: 500,
      body: `Error saving file: ${error.message}`,
    };
  }
}
```

发送上传文件请求：

```sh
curl --location --request PUT 'url?filename=file.png' \
    --header 'Content-Type: application/octet-stream' \
    --data 'file.png'
```

### 返回不同类型的响应

可综合使用集成响应、非集成响应，获得更丰富的响应类型：

```js
exports.main = function(event, context) {
  const {  httpContext } = context
  const { url } = httpContext
  const path = new URL(url).pathname

  // 根据访问路径返回不同的内容
  switch (path) {
    // 直接返回字符串
    case '/':
      return 'Hello world!'
    // 返回当前时间戳
    case '/now':
      return new Date().getTime()
    // 使用集成响应返回 HTML
    case '/index.html':
      return {
        statusCode: 200,
        headers: {
          'Content-Type': 'text/html'
        },
        body: '<h1>Hello world!</h1>'
      }
    // 使用集成响应返回 JSON
    default:
      return {
        statusCode: 404,
        headers: {
          'Content-Type': 'application/json'
        },
        body: {
          message: 'Not found'
        }
      }
  }
}
```
